//
//  Channel.cpp
//  e_bash
//
//  Created by Arnaud Blanchard on 29/09/2019.
//

#include "blc_channel.h"
#include "blc_image.h"
#include "ChannelColumns.hpp"
#include "ChannelWidget.hpp"
#include "Process.h"
#include "Command.hpp"

#include <numeric> //Accumulate
#include <string>
#include <vector>

using namespace Gtk;
using namespace Glib; //signal_idle,RefPtr
using  namespace std;

static void refresh_channels(void* user_data){
    auto channel_widget = (ChannelWidget *)user_data;
    signal_idle().connect([channel_widget](){ //important to use an idle otherwise Thread problems
        channel_widget->refresh_channel_list();
        return false;
    });
}

ChannelColumns ChannelWidget::columns;

ChannelWidget::ChannelWidget(){
    char const *blaar_dir=getenv("BLAAR_DIR");
    if (blaar_dir==nullptr) EXIT_ON_ERROR("Your environment variable 'BLAAR_DIR' is not defined.");
    bin_dir=string(blaar_dir)+"/bin";
    
    append_column("Name", columns.name);
    append_column("Type", columns.type);
    append_column("Format", columns.format);
    append_column("Dimensions", columns.dimensions);
    
    tree_store = TreeStore::create(columns);
    set_model(tree_store);
    set_activate_on_single_click();
    
    TreeViewColumn *col = new TreeViewColumn(" ", terminal_cell_renderer);
    col->add_attribute(terminal_cell_renderer, "icon-name",  columns.terminal_icon);
    append_column(*col);
    
    col =  new TreeViewColumn(" ", play_pause_cell_renderer);
    col->add_attribute(play_pause_cell_renderer, "icon-name",  columns.play_pause_icon);
    col->add_attribute(play_pause_cell_renderer, "sensitive", columns.play_pause_sensitive);
    append_column(*col);
    
    col =  new TreeViewColumn(" ", stop_cell_renderer);
    col->add_attribute(stop_cell_renderer, "icon-name",  columns.stop_icon);
    col->add_attribute(stop_cell_renderer, "sensitive", columns.stop_sensitive);
    append_column(*col);
    
    

    commands.emplace("graph", Command(bin_dir, "o_gnuplot", {}, [](vector<blc_channel const*> channels ){
        if (channels.size()>1) return false;
      }));
    commands.emplace("graph history", Command(bin_dir, "o_gnuplot", {"--history=100"}, [](vector<blc_channel const*> channels ){
         int total_length=channels[0]->total_length;
         return all_of(channels.begin()+1,channels.end(),[total_length](blc_channel const *channel){
             if (channel->dims_nb==1 && channel->total_length!=total_length) return false;
             else if (channel->dims_nb==2) return false;
             return true;
         });
       }));
    commands.emplace("keyboard", Command(bin_dir, "i_keyboard", {"--debug"}, [](vector<blc_channel const*> channels ){
        if (channels.size()==1 && channels[0]->total_length<=64) return true;
        else return false;
    }));

    
    signal_row_activated().connect([this](const TreeModel::Path& path, TreeViewColumn* column){
        int x, y;
        TreeIter iter=this->tree_store->get_iter(path);
        TreeModel::Row row = *iter;
        Process *process=row[columns.process];
        if (tree_store->iter_depth(iter) != 0){ //It is a process
            if (column->get_cell_position(terminal_cell_renderer, x, y)){//x, y are not used but it allow to chack the existanace of the cell renderer
                process->display_terminal();
            }
            else if (column->get_cell_position(play_pause_cell_renderer, x, y)){
                switch (process->status){
                    case Process::running:
                        process->pause();
                        row[columns.play_pause_icon]="media-playback-start";
                        break;
                    case Process::paused:
                        process->cont();
                        row[columns.play_pause_icon]="media-playback-pause";
                        break;
                    case Process::stopped:
                        process->run();
                        break;
                }
            }
            else if (column->get_cell_position(stop_cell_renderer, x, y)){
                process->stop();
                row [columns.stop_sensitive]=false;
            }
        }
    });
    
    signal_button_release_event().connect([this](GdkEventButton const *event){
        if  (event->button == 3){
            ContextMenu *menu=create_menu();
            menu->popup_at_pointer((GdkEvent*)event);
        }
        return true;
    });
    
    refresh_channel_list();
    blc_channel_check_for_event(refresh_channels, this);
    get_selection()->set_mode(SELECTION_MULTIPLE);
    show_all_children();
}

blc_channel *ChannelWidget::find_channel(int  id){
    blc_channel *channel= lower_bound(channels, channels+channels_nb, id, [](blc_channel const &a, int id){
        return (a.id < id);
    });
    return channel;
}

void ChannelWidget::after_launch(Process *process, std::vector<TreeModel::Path> rows){
    auto process_rows=make_shared<vector<TreeIter>>();
    string command = accumulate(process->arg_vector.begin()+1, process->arg_vector.end()-1, process->program_name, [](auto const &a, auto const &b){
        return a+" "+b;
    });
    
    for (auto const &path : rows){
        TreeNodeChildren node = tree_store->get_iter(path)->children();
        TreeIter row_iter=tree_store->append(node);
        process_rows->emplace_back(row_iter);
        TreeRow row=*row_iter;
        row[columns.name]=command;
        row[columns.terminal_icon]="utilities-terminal";
        row[columns.play_pause_icon]="media-playback-pause";
        row[columns.play_pause_sensitive]=true;
        row[columns.stop_icon]="media-playback-stop";
        row[columns.stop_sensitive]=true;
        row[columns.process]= process;
        expand_row(path, true);
    }
    
    process->on_quit([process_rows](int status){
        for (auto &row_iter : *process_rows){
            TreeRow row=*row_iter;
            if (status==0)  row[columns.terminal_icon]="process-stop";
            else row[columns.terminal_icon]="dialog-error";
            row[columns.play_pause_icon]="media-playback-start";
            row[columns.stop_sensitive]=false;
        }
    });
    show_all_children();
}

void ChannelWidget::on_quit(Process *process, int status, std::vector<TreeModel::Path> rows){
    for (auto const &path : rows){
        TreeRow row = *tree_store->get_iter(path);
        //  row[columns.name]=command;
        //   row[columns.terminal_icon]="utilities-terminal";
        if (status==0)  row[columns.terminal_icon]="process-stop";
        else  row[columns.terminal_icon]="dialog-error";
        row[columns.stop_icon]="media-playback-start";
        expand_row(path, true);
    }
}
/*
void ChannelWidget::create_and_execute_command(string program_name, initializer_list<const char*> arguments){
    commands.emplace_back(bin_dir, program_name, arguments));
    auto command=&commands.back();
    command->execute();
}*/

/*
void ContextMenu::add_program(char const *menu_name, std::string program_name, initializer_list<const char*> const optional_arguments, function <void(Process *process)> after_launch){
    auto item=new MenuItem(menu_name);
    auto args_vec=make_shared<vector<const char*>>(optional_arguments); //This will be destroyed by process
    append(*item);
    
    item->signal_activate().connect([this, args_vec, program_name, after_launch]{
        args_vec->reserve(args_vec->size()+selected_channels.size());
        for(auto const &channel:selected_channels){
            args_vec->push_back(strdup(channel->name));
        }
        auto process = new Process(program_dir, program_name, args_vec);
        process->run(after_launch);
    });
}*/


/*
item->signal_activate().connect([this, args_vec, program_name, after_launch]{
    args_vec->reserve(args_vec->size()+selected_channels.size());
    for(auto const &channel:selected_channels){
        args_vec->push_back(strdup(channel->name));
    }
    auto process = new Process(program_dir, program_name, args_vec);
    process->run(after_launch);
});
 */

ContextMenu *ChannelWidget::create_menu(){
    bool only_channels=true;
    vector<Command*> context_commands;
    vector<blc_channel const *> selected_channels;
    menu=new ContextMenu();
    menu->row_selection=get_selection();
    
    get_selection()->selected_foreach_iter ([&](TreeModel::iterator const &iter){
           if (tree_store->iter_depth(iter)==0){
               blc_channel *channel=find_channel((*iter)[columns.id]);
               selected_channels.emplace_back(channel);
           }
           else only_channels=false;
       });
    
    if (only_channels){
        menu->add_item("_Remove", true, [this]{
            for( auto &channel : menu->selected_channels){
                channel->remove();
            }
        });
        menu->add_separator();
        
        for ( const auto&  [name, command]:commands){
            Command const *command_ptr=&command; //wihout ptr, command cannot be pass to lambda
            if (command.valid_channels(selected_channels));
            menu->add_item(name, false, [this, command_ptr, selected_channels](){
                auto command =used_commands.emplace_back(*command_ptr, selected_channels);
                auto process=new Process(command.args.data(), [](Process*){});
            });
        }
    }
    menu->show_all();
    return menu;
}

/**
 We look for all channels. We get a list of actual row and look for differences to add or remove channel in the known list.
 This suppose that a channel with a same id has the same properties. It is a property of blc_channel.
 */
void ChannelWidget::refresh_channel_list(){
    ///  We define a functor to compare blc_channel and TreeIter by id
    struct compare_id{
        bool operator()(blc_channel const &a, TreeIter const &b) { return a.id < (*b)[columns.id]; };
            bool operator()(TreeIter const &a, blc_channel const &b) { return  (*a)[columns.id]< b.id ; };
            };
            vector<blc_channel> news_channels;
            vector<TreeIter> channels_to_remove;
            
            FREE(channels);
            tie(channels, channels_nb)=blc_channel::get_all_infos();
            
            //We get children to be able to use iterator
            auto children=tree_store->children();
            
            //We get a list of new channels
            set_difference(channels, channels+channels_nb, children.begin(), children.end(), inserter(news_channels, news_channels.begin()), compare_id());
            
            //We get a list of channels to remove
            set_difference(children.begin(), children.end(), channels, channels+channels_nb, inserter(channels_to_remove, channels_to_remove.begin()), compare_id());
            
            for(auto const &tree_iter : channels_to_remove)  tree_store->erase(tree_iter);
            for(auto const &channel :news_channels)  columns.add_channel(tree_store, channel);
            
            this->show_all_children();
            
            //We free the list of channel before getting a new list. This is not optimum but it does not happen often.
            for_each(channels, channels+channels_nb, [](blc_channel channel){
                channel.~blc_channel(); //We destroy the dims inside
            });
            }
